#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "objz80.h"

extern int prefix;

struct z80fileinfo ofi;

unsigned char * workspace = NULL;
unsigned short pc = 0;
int line = 0;

struct opt { char * operand; void (*fn)(char *); } operand_table[]={
	    "word ", assemble_word, 
	    "dsbyte ", assemble_dsbyte, 
	    "byte ", assemble_byte, 
	    "ld ", assemble_ld,
	    "pop ", assemble_pop,
	    "push ", assemble_push,
	    "ex ", assemble_ex,
	    "exx", assemble_exx,

	    "add ", assemble_add,
	    "adc ", assemble_adc,
	    "sbc ", assemble_sbc,
	    "sub ", assemble_sub,
	    "cp ", assemble_cp,
	    "and ", assemble_and,
	    "or ", assemble_or,
	    "xor ", assemble_xor,
	    "inc ", assemble_inc,
	    "dec ", assemble_dec,
	    "cpl", assemble_cpl,
	    "neg", assemble_neg,
	    "scf", assemble_scf,
	    "ccf", assemble_ccf,
	    "daa", assemble_daa,

	    "rlca", assemble_rlca,
	    "rrca", assemble_rrca,
	    "rla", assemble_rla,
	    "rra", assemble_rra,
	    "rlc ", assemble_rlc,
	    "rrc ", assemble_rrc,
	    "rl ", assemble_rl,
	    "rr ", assemble_rr,
	    "sla ", assemble_sla,
	    "sra ", assemble_sra,
	    "sli ", assemble_sli,
	    "srl ", assemble_srl,

	    "rrd", assemble_rrd,
	    "rld", assemble_rld,

	    "ldir", assemble_ldir,
	    "lddr", assemble_lddr,
	    "cpir", assemble_cpir,
	    "cpdr", assemble_cpdr,
	    "ldi", assemble_ldi,
	    "ldd", assemble_ldd,
	    "nop", assemble_nop,
	    "halt", assemble_halt,
	    "di", assemble_di,
	    "ei", assemble_ei,
	    "im ", assemble_im,
	    "jp ", assemble_jp,
	    "jr ", assemble_jr,
	    "call ", assemble_call,
	    "djnz ", assemble_djnz,
	    "reti", assemble_reti,
	    "retn", assemble_retn,
	    "ret", assemble_ret,
	    "rst ", assemble_rst,
	    "bit ", assemble_bit,
	    "set ", assemble_set,
	    "res ", assemble_res,

	    "in ", assemble_in,
	    "out ", assemble_out,
	    "ini ", assemble_ini,
	    "ind ", assemble_ind,
	    "outi", assemble_outi,
	    "outd", assemble_outd,
	    "inir", assemble_inir,
	    "indr", assemble_indr,
	    "otir", assemble_otir,
	    "otdr", assemble_otdr,
	    NULL};

void putbyte(unsigned char b)
{
    workspace[pc++] = b;
}

void putword(unsigned short w)
{
    workspace[pc++] = w & 0xff;
    workspace[pc++] = (w >> 8) & 0xff;
}
void singleasm(char * line)
{
    int i;
    unsigned char * opdata;
#ifdef DEBUG
    printf("requested assembly of '%s'\n", line);
#endif
    prefix=0x00;
    for(i=0; operand_table[i].operand; i++)
	if(strncmp(line, operand_table[i].operand, strlen(operand_table[i].operand))==0)
	{
	    operand_table[i].fn(line+strlen(operand_table[i].operand));
	    break;
	}
    if(operand_table[i].operand == NULL) fprintf(stderr,"Warning, unknown instruction in '%s'\n",line);
#ifdef DEBUG
    printf("ok, done\n");
#endif
}

void process_directive(unsigned char *item)
{
    if(strncmp(item,"!global ",8) == 0)
    {
	make_global(item+8);
    }
}

void assemble(unsigned char * mem, int len, int outfd)
{
    int i;
    
    workspace = (unsigned char *) malloc(65536);
    for(;;)
    {
	if(len<0) break;
	if(*mem=='\n') { line ++; mem++; len--; continue; }
	/* skip comments */
	if(*mem == '#' || *mem==';')
	{
	    for(i=0; i<len && mem[i]!='\n'; i++);
	    if(i == len) break;
	    mem+=i+1; len-=i+1;
	    line ++;
	    continue;
	}
	if(*mem=='!')
	{
	    for(i=0; i<len && mem[i]!='\n'; i++);
	    if(i == len) break;
	    mem[i] = 0;
	    process_directive(mem);
	    mem+=i+1; len-=i+1;
	    line ++;
	    continue;
	}
	if(isblank(*mem)) /* operand */
	{
	    unsigned char * com;
	    /* skip blanks */
	    for(i=0; i<len && isblank(mem[i]); i++);
	    if(i == len) break;
	    mem+=i; len-=i;
	    /* find EOL */
	    for(i=0; i<len && mem[i]!='\n'; i++);
	    if(i == len) break;
	    mem[i] = 0;
	    if(com = index(mem,';')) *com = 0;
	    if(strlen(mem))
	    {
		int q;

		for(q = strlen(mem) - 1; q>=0; q--)
		    if(isblank(mem[q])) mem[q] = 0; else break;
	    }
	    
	    if(strlen(mem)) singleasm(mem);
	    mem+=i+1; len-=i+1;
	    line++;
	}
	else	/* not blank -> label */
	{
	    /* find first blank */
	    for(i=0; i<len && !isblank(mem[i]) && mem[i]!='\n'; i++);
	    if(i == len) break;
	    if(mem[i]=='\n') line++;
	    mem[i] = 0;
	    /* mem is the label */
	    add_symbol(mem, pc);
	    mem+=i+1; len-=i+1;
	}
	
    }
    complete_symbol_values();
    save(outfd, pc);
}
