#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "objz80.h"

unsigned char * load(char * filename, int * slen)
{
    int fd;
    int len;
    unsigned char * ret;
    
    if((fd = open(filename, O_RDONLY))==-1) return(NULL);
    if((len = lseek(fd, 0, SEEK_END)) == -1) { perror("lseek"); close(fd); return(NULL); }
    if(lseek(fd, 0, SEEK_SET) == -1) { perror("lseek"); close(fd); return(NULL); }
    
    ret = (unsigned char *) malloc(len);
    if(read(fd, ret, len)!=len) { perror("read"); close(fd); return(NULL); }
    close(fd);
    *slen = len;
    return(ret);
}

int ckreg8(char * reg)
{
    unsigned char d;
    if(strcmp(reg,"a")==0)  return(7);
    if(strcmp(reg,"b")==0)  return(0);
    if(strcmp(reg,"c")==0)  return(1);
    if(strcmp(reg,"d")==0)  return(2);
    if(strcmp(reg,"e")==0)  return(3);
    if(strcmp(reg,"h")==0)  return(4);
    if(strcmp(reg,"l")==0)  return(5);
    if(strcmp(reg,"(hl)")==0)  return(6);
}

int b2d(char * str)
{
    int d = 0;
    for(;;)
    {
	if(*str == '1')
	{
	    d  = d <<1;
	    d = d | 1;
	}
	else if(*str == '0')
	{
	    d  = d <<1;
	}
	else break;
	str++;
    }
    return(d);
}

void parsearg(char *name, struct arg * arg)
{
    arg->str=name;
    if(strcmp(name,"a")==0 || strcmp(name,"b")==0 || strcmp(name,"c")==0 || strcmp(name,"d")==0 ||
       strcmp(name,"e")==0 || strcmp(name,"h")==0 || strcmp(name,"l")==0 || strcmp(name,"(hl)")==0)
    {
	arg->type = AT_8BITREG; arg->prefixed = 0; arg->relocate = 0; arg->offset = 0;
	arg->value = ckreg8(name);
	return;
    }
    if(strncmp(name,"(ix",3) == 0)
    {
	arg->type = AT_8BITREG; arg->prefixed = 0xdd; arg->relocate = 0; arg->offset = (strtol(name+3, NULL, 0) & 0xff);
	arg->value = 6;
	return;
    }
    if(strncmp(name,"(iy",3) == 0)
    {
	arg->type = AT_8BITREG; arg->prefixed = 0xfd; arg->relocate = 0; arg->offset = (strtol(name+3, NULL, 0) & 0xff);
	arg->value = 6;
	return;
    }
    if(strcmp(name,"bc") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0; arg->relocate = 0; arg->offset = 0;
	arg->value = 0;
	return;
    }

    if(strcmp(name,"de") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0; arg->relocate = 0; arg->offset = 0;
	arg->value = 1;
	return;
    }
    if(strcmp(name,"hl") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0; arg->relocate = 0; arg->offset = 0;
	arg->value = 2;
	return;
    }
    if(strcmp(name,"ix") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0xdd; arg->relocate = 0; arg->offset = 257;
	arg->value = 2;
	return;
    }
    if(strcmp(name,"iy") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0xfd; arg->relocate = 0; arg->offset = 257;
	arg->value = 2;
	return;
    }
    if(strcmp(name,"sp") == 0 || strcmp(name,"af") == 0)
    {
	arg->type = AT_16BITREG; arg->prefixed = 0; arg->relocate = 0; arg->offset = 0;
	arg->value = 3;
	return;
    }
    
    if(*name=='(')
    {
	name++;
	if(*name == '#')
	{
	    arg->relocate = 0; name++;
	}
	else
	    arg->relocate = 1;
	if(isdigit(*name))
	{
	    arg->type = AT_MEMPTR; arg->prefixed = 0; arg->offset = 0; arg->value = strtol(name, NULL, 0);
	}
	else if(*name == '%')
	{
	    arg->type = AT_MEMPTR; arg->prefixed = 0; arg->offset = 0; arg->value = b2d(name + 1);
	}
	else
	{
	    char *q;
	    
	    arg->type = AT_SYMBOLIC | AT_MEMPTR; arg->prefixed = 0; arg->offset = 0; 
	    arg->str = name;
	    if(q = index(name, ')')) *q=0;
	}
	return;
    }

    if(*name == '#')
    {
	arg->relocate = 0; name ++;
    }
    else
	arg->relocate = 1;

    if(isdigit(*name))
    {
	arg->type = AT_NUMCONST; arg->prefixed = 0; arg->offset = 0; arg->value = strtol(name, NULL, 0);
	return;
    }
    else if(*name == '%')
    {
	arg->type = AT_NUMCONST; arg->prefixed = 0; arg->offset = 0; arg->value = b2d(name+1);
	return;
    }
    else
    {
	arg->type = AT_16BITCONST | AT_SYMBOLIC; arg->prefixed = 0; arg->offset = 0;
	return;
    }
}
